from Bio import SeqIO, GenBank
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import argparse

def revcomp(seq):
    pairs = {'A':'T', 'T':'A', 'G':'C', 'C':'G'}
    return ''.join(list(pairs[i] for i in seq[::-1]))

part_type_info = \
    {'extra': {'reases5':'GAAGACTCGAGTCGTCTCATCGGTCTCATATG', 'reases3':'GAAGACGCAGGGCGTCTCTAGAA'},
     'spacer': {'reases5':'GAAGACTCGAGTCGTCTCATTCT', 'reases3':'GAAGACGCAGGGCGTCTCTCGTT'},
     'trans': {'reases5':'GAAGACTCGAGTCGTCTCAAACG', 'reases3':'GAAGACGCAGGGCGTCTCTATTG'},
     'intra': {'reases5':'GAAGACTCGAGTCGTCTCACAAT', 'reases3':'GAAGACGCAGGGCGTCTCTGGTCGGTCTCAGGAT'}}

def make_primers(seq, **part_type_info):
    ''' 
    takes a DNA sequence and returns upstream/downstream primers to PCR 
    out an extracellular CAR-TK part
     
    requires that part_type_info contains `reases5` and `reases3`, which are the 
    regions of the primer characteristic of the CAR-TK part type (e.g. extra vs intra) '''

    clamp5 = 'GCAT'  # upstream primer clamp
    clamp3 = 'ATGC'  # downstream primer clamp
    beginning = seq.upper()[:25]

    # try to end in G or C
    while beginning[-1] in ['A', 'T'] and len(beginning) >= 20:
        beginning = beginning[:-1]

    end = revcomp(seq.upper()[-25:])

    # try to end in G or C
    while end[-1] in ['A', 'T'] and len(end) >= 20:
        end = end[:-1]

    return {'upstream':clamp5 + part_type_info['reases5'] + beginning,
            'downstream':clamp3 + part_type_info['reases3'] + end}


def primer(plasmid_path, input_dict, save_path=None):
    ''' 
    plasmid: path to gb file with annotations 
    input_dict: dictionary whose keys are annotations (e.g. `antiIL6 scFv`)
    and whose values are the desired CAR-TK type
    '''
    import IPython

    assert plasmid_path[-2:] == 'gb', 'Please upload a GenBank (.gb) file'

    with open(plasmid_path) as file:
        record = list(SeqIO.parse(file, 'genbank'))[0]
        #IPython.embed(colors='neutral')

    # convert plasmid's goofy format to dictionary annotation : sequence

    plasmid = {}
    annotations = list(dict(record.features[i].qualifiers.items())['label'][0] for i in range(len(record.features)))
    assert set(input_dict.keys()).issubset(set(annotations)), f'One or more of your entered annotations ({", ".join(sorted(set(input_dict.keys()).difference(set(annotations))))}) does not exist in your plasmid.'

    for i, annotation in enumerate(annotations):
        plasmid[annotation] = (int(record.features[i].location.start),
                               int(record.features[i].location.end),
                               record.features[i].location.strand)
    
    PLASMID_SEQ = str(record.seq)  # entire plasmid's sequence
    primers = {}
    for annotation in input_dict.keys():
        seq = PLASMID_SEQ[plasmid[annotation][0]:plasmid[annotation][1]] if plasmid[annotation][2] == 1 \
        else revcomp(PLASMID_SEQ[plasmid[annotation][0]:plasmid[annotation][1]])

        primers[annotation] = make_primers(seq,
            reases5=part_type_info[input_dict[annotation]]['reases5'],
            reases3=part_type_info[input_dict[annotation]]['reases3'])
        
    if save_path:
        for annotation in primers.keys():
            # create genbank objects
            for strand in ['upstream', 'downstream']:
                gb = Seq(primers[annotation][strand])

                # create SeqRecord object
                record = SeqRecord(gb, name = f"{annotation.replace(' ', '_')}-{strand}-primer")
                record.annotations["molecule_type"] = "DNA"
                # Save as GenBank file
                with open(f"{save_path}{annotation}_{strand}_primer.gb", 'w') as output_file:
                    SeqIO.write(record, output_file, 'genbank')
        
    return primers
    

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--plasmid_path')
    parser.add_argument('-s', '--save-path')
    parser.add_argument('-a', '--annotations', nargs='*')
    args = parser.parse_args()

    cartk_parts = ['extra', 'spacer', 'trans', 'intra']
    assert len(args.annotations) % 2 == 0, 'You are missing an annotation argument.'
    
    for i in range(0,len(args.annotations),2):
        assert args.annotations[i+1] in cartk_parts, f"{args.annotations[i+1]} is not a recognized CAR-TK part type."
   
    annotations_dict = {args.annotations[i]:args.annotations[i+1] for \
                        i in range(0, len(args.annotations), 2)}

    print(primer(args.plasmid_path, annotations_dict, save_path=args.save_path))
    
    # so u have to make sure you escape spaces when you enter stuff 
    # into the command line, e.g. `GS\ linker\ (paper)`

    # {'CD8\ signal\ and\ anti\ IL-6' : 'extra',
    # 'GS\ linker\ (paper)' : 'spacer',
    # 'CD8\ transmembrane\ stalk\ (aa\ 138-206;\ Q96QR6_HUMAN)' : 'trans',
    # 'Fc\ epsilon\ receptor\ gamma\ subunit\ (aa\ 45-86;\ FCERG_MOUSE)' : 'intra'}
