from Bio.Restriction import RestrictionBatch
from Bio.Seq import Seq
from Bio import SeqIO

from itertools import chain, combinations
import numpy as np
import pathlib
import argparse

def read_enzymes(fname):
    pass

def read_plasmid(fname):
    fext = pathlib.Path(fname).suffix.lstrip(".")

    ext = None

    if fext in {"ape", "gb", "genbank", "gbk"}:
        ext = "gb"
    elif fext in {"fasta"}:
        ext = "fasta"
    
    if not ext:
        raise Exception("unknown file type")

    plasmid = SeqIO.read(fname, ext)

    return plasmid

def compute_fragment_lengths(plasmid_len, cut_indices):
    """
        returns: a tuple where the first element is a list of fragment lengths, and the second is a boolean indicating whether or not the plasmid remains supercoiled. 
    """
    if len(cut_indices) == 0:
        return [plasmid_len], True

    cut_indices = sorted(cut_indices)
    cut_indices.append(plasmid_len + cut_indices[0])

    fragment_lengths = []

    for i in range(len(cut_indices) - 1):
        fragment_lengths.append(cut_indices[i + 1] - cut_indices[i])
    
    return sorted(fragment_lengths), False

def convert_fragments_to_distance(fragments):
    bp = (10000, 100)
    fitdists = (1, 50)
    # 10 kb = 0 mm, 0.1 kb = 50 mm

    # y = be^(ax)
    # ln(y) = ax + ln(b)
    # - x is fragment length
    # - y is distance travelled
    fit = np.polyfit(bp, np.log(fitdists), 1)
    lndists = np.polyval(fit, fragments)
    dists = np.exp(lndists)

    return list(dists)

class Digest():
    def __init__(self, enzymes, fragments, distances):
        self.enzymes = enzymes
        self.fragments = sorted(fragments)
        self.distances = sorted(distances) # in mm
    
    def score_digest(self):
        bad = 0

        for fragment in self.fragments:
            if fragment < 500:
                bad += 1
        
        for i in range(len(self.distances) - 1):
            dist = self.distances[i + 1] - self.distances[i]
            if dist < 5:
                bad += 1
        
        if len(self.enzymes) > 1:
            bad += 1
        
        return bad
    
    def __str__(self):
        return "\n".join([
            ",".join(self.enzymes),
            ",".join(map(str, self.fragments))
        ])
    
    def num_cuts(self):
        return len(self.fragments)
    
    def to_dict(self):
        return {
            "fragments": self.fragments,
            "enzymes": self.enzymes,
            "score": self.score_digest()
        }


def rank_digests(digests):
    scored_digests = [d.to_dict() for d in digests]
    ranked_digests = sorted(scored_digests, key=lambda x: x["score"])

    return ranked_digests

def design_diagnostic_digest(seq, enzymes):
    seq = Seq(seq)

    combos = chain(combinations(enzymes, 1), combinations(enzymes, 2))

    digests = []

    # cut
    for enz in combos:
        rb = RestrictionBatch(enz)
        scan = rb.search(seq, linear=False)
        cutsites = set()

        skip = False
        for _, sites in scan.items():
            if len(sites) == 0:
                skip = True
                break
            cutsites |= set(sites)
        if skip:
            continue
        
        fragments, supercoiled = compute_fragment_lengths(len(seq), list(cutsites))

        distances = convert_fragments_to_distance(fragments)

        digest = Digest(
            enz,
            fragments,
            distances
        )

        if not supercoiled:
            digests.append(digest)
    
    single = filter(lambda x: x.num_cuts() == 1, digests)
    multi = filter(lambda x: x.num_cuts() > 1, digests)
    
    ranked_single = rank_digests(single)
    ranked_multi = rank_digests(multi)

    return ranked_single, ranked_multi


def main(plasmid_fname):
    enzymes = ["ApaLI", "AseI", "BamHI", "BsaAI", "EcoRI", "EcoRV"]
    plasmid = read_plasmid(plasmid_fname)

    ranked_single, ranked_multi = design_diagnostic_digest(str(plasmid.seq), enzymes)

    for digest, score in ranked_single:
        print("-" * 24)
        print(score)
        print(digest)
    
    print("#" * 24)

    for digest, score in ranked_multi:
        print(score)
        print(digest)
        print("-" * 24)



if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # parser.add_argument("enzymes", help="CSV of enzymes and their associated properties (including cut sites, star activity, incubation temperature, etc.)")
    parser.add_argument("plasmids", help="plasmid sequences in FASTA format")

    args = parser.parse_args()

    main(args.plasmids)
